//  Copyright (c) 2017-2018 John Biddiscombe
//
//  SPDX-License-Identifier: BSL-1.0
//  Distributed under the Boost Software License, Version 1.0. (See accompanying
//  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#include <hpx/include/parallel_executors.hpp>
#include <hpx/runtime/threads/executors/default_executor.hpp>
#include <hpx/runtime/threads/executors/pool_executor.hpp>
#define GUIDED_EXECUTOR_DEBUG 1
#include <hpx/resource_partitioner/partitioner.hpp>
#include <hpx/runtime/threads/executors/guided_pool_executor.hpp>
//#include <hpx/topology/cpu_mask.hpp>
//#include <hpx/include/parallel_executors.hpp>
#include <hpx/async.hpp>

// we should not need these
#include <hpx/runtime/threads/detail/scheduled_thread_pool_impl.hpp>

#include <hpx/datastructures/tuple.hpp>
#include <hpx/debugging/demangle_helper.hpp>
#include <hpx/functional/deferred_call.hpp>
#include <hpx/functional/invoke.hpp>
#include <hpx/functional/invoke_fused.hpp>
#include <hpx/functional/result_of.hpp>
#include <hpx/lcos/dataflow.hpp>
#include <hpx/lcos/when_all.hpp>
#include <hpx/type_support/decay.hpp>
#include <hpx/util/pack_traversal.hpp>
//
#include <cstddef>
#include <cstdint>
#include <iostream>
#include <memory>
#include <string>
#include <type_traits>
#include <utility>

// --------------------------------------------------------------------
// custom executor async/then/when/dataflow specialization example
// --------------------------------------------------------------------
using namespace hpx;

struct test_async_executor
{
    // --------------------------------------------------------------------
    // helper structs to make future<tuple<f1, f2, f3, ...>>>
    // detection of futures simpler
    // --------------------------------------------------------------------
    template <typename TupleOfFutures>
    struct is_tuple_of_futures;

    template <typename... Futures>
    struct is_tuple_of_futures<util::tuple<Futures...>>
      : util::detail::all_of<
            traits::is_future<typename std::remove_reference<Futures>::type>...>
    {
    };

    template <typename Future>
    struct is_future_of_tuple_of_futures
      : std::integral_constant<bool,
            traits::is_future<Future>::value &&
                is_tuple_of_futures<typename traits::future_traits<
                    typename std::remove_reference<Future>::type>::
                        result_type>::value>
    {
    };

    // --------------------------------------------------------------------
    // For C++11 compatibility
    template <bool B, typename T = void>
    using enable_if_t = typename std::enable_if<B, T>::type;

    // --------------------------------------------------------------------
    // function that returns a const ref to the contents of a future
    // without calling .get() on the future so that we can use the value
    // and then pass the original future on to the intended destination.
    // --------------------------------------------------------------------
    struct future_extract_value
    {
        template <typename T, template <typename> class Future>
        const T& operator()(const Future<T>& el) const
        {
            typedef
                typename traits::detail::shared_state_ptr_for<Future<T>>::type
                    shared_state_ptr;
            shared_state_ptr const& state =
                traits::detail::get_shared_state(el);
            return *state->get_result();
        }
    };

    // --------------------------------------------------------------------
    // async execute specialized for simple arguments typical
    // of a normal async call with arbitrary arguments
    // --------------------------------------------------------------------
    template <typename F, typename... Ts>
    future<typename util::invoke_result<F, Ts...>::type> async_execute(
        F&& f, Ts&&... ts)
    {
        typedef typename util::detail::invoke_deferred_result<F, Ts...>::type
            result_type;

        using namespace hpx::util::debug;
        std::cout << "async_execute : Function    : " << print_type<F>()
                  << "\n";
        std::cout << "async_execute : Arguments   : "
                  << print_type<Ts...>(" | ") << "\n";
        std::cout << "async_execute : Result      : "
                  << print_type<result_type>() << "\n";

        // forward the task execution on to the real internal executor
        lcos::local::futures_factory<result_type()> p(executor_,
            util::deferred_call(std::forward<F>(f), std::forward<Ts>(ts)...));

        p.apply("custom", launch::async, threads::thread_priority_default,
            threads::thread_stacksize_default);

        return p.get_future();
    }

    // --------------------------------------------------------------------
    // .then() execute specialized for a future<P> predecessor argument
    // note that future<> and shared_future<> are both supported
    // --------------------------------------------------------------------
    template <typename F, typename Future, typename... Ts,
        typename = enable_if_t<traits::is_future<
            typename std::remove_reference<Future>::type>::value>>
    auto then_execute(F&& f, Future&& predecessor, Ts&&... ts) -> future<
        typename util::detail::invoke_deferred_result<F, Future, Ts...>::type>
    {
        typedef typename util::detail::invoke_deferred_result<F, Future,
            Ts...>::type result_type;

        using namespace hpx::util::debug;
        std::cout << "then_execute : Function     : " << print_type<F>()
                  << "\n";
        std::cout << "then_execute : Predecessor  : " << print_type<Future>()
                  << "\n";
        std::cout
            << "then_execute : Future       : "
            << print_type<typename traits::future_traits<Future>::result_type>()
            << "\n";
        std::cout << "then_execute : Arguments    : "
                  << print_type<Ts...>(" | ") << "\n";
        std::cout << "then_execute : Result       : "
                  << print_type<result_type>() << "\n";

        // forward the task on to the 'real' underlying executor
        lcos::local::futures_factory<result_type()> p(executor_,
            util::deferred_call(std::forward<F>(f),
                std::forward<Future>(predecessor), std::forward<Ts>(ts)...));

        p.apply("custom then", launch::async, threads::thread_priority_default,
            threads::thread_stacksize_default);

        return p.get_future();
    }

    // --------------------------------------------------------------------
    // .then() execute specialized for a when_all dispatch for any future types
    // future< tuple< is_future<a>::type, is_future<b>::type, ...> >
    // --------------------------------------------------------------------
    template <typename F, template <typename> class OuterFuture,
        typename... InnerFutures, typename... Ts,
        typename = enable_if_t<is_future_of_tuple_of_futures<
            OuterFuture<util::tuple<InnerFutures...>>>::value>,
        typename = enable_if_t<
            is_tuple_of_futures<util::tuple<InnerFutures...>>::value>>
    auto then_execute(F&& f,
        OuterFuture<util::tuple<InnerFutures...>>&& predecessor, Ts&&... ts)
        -> future<typename util::detail::invoke_deferred_result<F,
            OuterFuture<util::tuple<InnerFutures...>>, Ts...>::type>
    {
        typedef typename util::detail::invoke_deferred_result<F,
            OuterFuture<util::tuple<InnerFutures...>>, Ts...>::type result_type;

        // get the tuple of futures from the predecessor future <tuple of futures>
        const auto& predecessor_value =
            future_extract_value().operator()(predecessor);

        // create a tuple of the unwrapped future values
        auto unwrapped_futures_tuple =
            util::map_pack(future_extract_value{}, predecessor_value);

        using namespace hpx::util::debug;
        std::cout << "when_all(fut) : Predecessor : "
                  << print_type<OuterFuture<util::tuple<InnerFutures...>>>()
                  << "\n";
        std::cout << "when_all(fut) : unwrapped   : "
                  << print_type<decltype(unwrapped_futures_tuple)>(" | ")
                  << "\n";
        std::cout << "when_all(fut) : Arguments   : "
                  << print_type<Ts...>(" | ") << "\n";
        std::cout << "when_all(fut) : Result      : "
                  << print_type<result_type>() << "\n";

        // invoke a function with the unwrapped tuple future types to demonstrate
        // that we can access them
        std::cout << "when_all(fut) : tuple       : ";
        util::invoke_fused(
            [](const auto&... ts) {
                std::cout << print_type<decltype(ts)...>(" | ") << "\n";
            },
            unwrapped_futures_tuple);

        // forward the task execution on to the real internal executor
        lcos::local::futures_factory<result_type()> p(executor_,
            util::deferred_call(std::forward<F>(f),
                std::forward<OuterFuture<util::tuple<InnerFutures...>>>(
                    predecessor),
                std::forward<Ts>(ts)...));

        p.apply("custom then", launch::async, threads::thread_priority_default,
            threads::thread_stacksize_default);

        return p.get_future();
    }

    // --------------------------------------------------------------------
    // execute specialized for a dataflow dispatch
    // dataflow unwraps the outer future for us but passes a dataflowframe
    // function type, result type and tuple of futures as arguments
    // --------------------------------------------------------------------
    template <typename F, typename... InnerFutures,
        typename = enable_if_t<
            traits::is_future_tuple<util::tuple<InnerFutures...>>::value>>
    auto async_execute(F&& f, util::tuple<InnerFutures...>&& predecessor)
        -> future<typename util::detail::invoke_deferred_result<F,
            util::tuple<InnerFutures...>>::type>
    {
        typedef typename util::detail::invoke_deferred_result<F,
            util::tuple<InnerFutures...>>::type result_type;

        auto unwrapped_futures_tuple =
            util::map_pack(future_extract_value{}, predecessor);

        using namespace hpx::util::debug;
        std::cout << "dataflow      : Predecessor : "
                  << print_type<util::tuple<InnerFutures...>>() << "\n";
        std::cout << "dataflow      : unwrapped   : "
                  << print_type<decltype(unwrapped_futures_tuple)>(" | ")
                  << "\n";
        std::cout << "dataflow-frame: Result      : "
                  << print_type<result_type>() << "\n";

        // invoke a function with the unwrapped tuple future types to demonstrate
        // that we can access them
        std::cout << "dataflow      : tuple       : ";
        util::invoke_fused(
            [](const auto&... ts) {
                std::cout << print_type<decltype(ts)...>(" | ") << "\n";
            },
            unwrapped_futures_tuple);

        // forward the task execution on to the real internal executor
        // forward the task execution on to the real internal executor
        lcos::local::futures_factory<result_type()> p(
            util::deferred_call(std::forward<F>(f),
                std::forward<util::tuple<InnerFutures...>>(predecessor)));

        p.apply("custom async", launch::async, threads::thread_priority_default,
            threads::thread_stacksize_default);

        return p.get_future();
    }

private:
    threads::executors::default_executor executor_;
};

// --------------------------------------------------------------------
// set traits for executor to say it is an async executor
// --------------------------------------------------------------------
namespace hpx { namespace parallel { namespace execution {
    template <>
    struct is_two_way_executor<test_async_executor> : std::true_type
    {
    };
}}}    // namespace hpx::parallel::execution

// --------------------------------------------------------------------
// test various execution modes
// --------------------------------------------------------------------
template <typename Executor>
int test(const std::string& message, Executor& exec)
{
    // test 1
    std::cout << "============================" << std::endl;
    std::cout << message << std::endl;
    std::cout << "============================" << std::endl;
    std::cout << "Test 1 : async()" << std::endl;
    auto fa = async(
        exec,
        [](int a, double b, const char* c) {
            std::cout << "Inside async " << c << std::endl;
            return float(a * b) + 2.1415f;
        },
        1, 2.2, "Hello");
    fa.get();
    std::cout << std::endl;

    // test 2a
    std::cout << "============================" << std::endl;
    std::cout << "Test 2a : .then()" << std::endl;
    future<int> f = make_ready_future(5);
    //
    future<std::string> ft = f.then(exec, [](future<int>&& /*f*/) {
        std::cout << "Inside .then()" << std::endl;
        return std::string("then");
    });
    ft.get();
    std::cout << std::endl;

    // test 2b
    std::cout << "============================" << std::endl;
    std::cout << "Test 2b : .then(shared)" << std::endl;
    auto fs = make_ready_future(5).share();
    //
    future<std::string> fts = fs.then(exec, [](shared_future<int>&& /*f*/) {
        std::cout << "Inside .then(shared)" << std::endl;
        return std::string("then(shared)");
    });
    fts.get();
    std::cout << std::endl;

    // test 3a
    std::cout << "============================" << std::endl;
    std::cout << "Test 3a : when_all()" << std::endl;
    future<int> fw1 = make_ready_future(123);
    future<double> fw2 = make_ready_future(4.567);
    //
    auto fw = when_all(fw1, fw2).then(
        exec, [](future<util::tuple<future<int>, future<double>>>&& f) {
            auto tup = f.get();
            auto cmplx = std::complex<double>(
                double(util::get<0>(tup).get()), util::get<1>(tup).get());
            std::cout << "Inside when_all : " << cmplx << std::endl;
            return std::string("when_all");
        });
    fw.get();
    std::cout << std::endl;

    // test 3b
    std::cout << "============================" << std::endl;
    std::cout << "Test 3b : when_all(shared)" << std::endl;
    future<std::uint64_t> fws1 = make_ready_future(std::uint64_t(42));
    shared_future<float> fws2 = make_ready_future(3.1415f).share();
    //
    auto fws = when_all(fws1, fws2)
                   .then(exec,
                       [](future<util::tuple<future<std::uint64_t>,
                               shared_future<float>>>&& f) {
                           auto tup = f.get();
                           auto cmplx = std::complex<double>(
                               double(util::get<0>(tup).get()),
                               double(util::get<1>(tup).get()));
                           std::cout << "Inside when_all(shared) : " << cmplx
                                     << std::endl;
                           return cmplx;
                       });
    fws.get();
    std::cout << std::endl;

    // test 4a
    std::cout << "============================" << std::endl;
    std::cout << "Test 4a : dataflow()" << std::endl;
    future<std::uint16_t> f1 = make_ready_future(std::uint16_t(255));
    future<double> f2 = make_ready_future(127.890);
    //
    auto fd = dataflow(
        exec,
        [](future<std::uint16_t>&& f1, future<double>&& f2) {
            auto cmplx = std::complex<std::uint64_t>(
                f1.get(), static_cast<std::uint64_t>(f2.get()));
            std::cout << "Inside dataflow : " << cmplx << std::endl;
            return cmplx;
        },
        f1, f2);
    fd.get();
    std::cout << std::endl;

    // test 4b
    std::cout << "============================" << std::endl;
    std::cout << "Test 4b : dataflow(shared)" << std::endl;
    future<std::uint32_t> fs1 = make_ready_future(std::uint32_t(65535));
    shared_future<double> fs2 = make_ready_future(2.178).share();
    //
    auto fds = dataflow(
        exec,
        [](future<std::uint32_t>&& f1, shared_future<double>&& f2) {
            auto cmplx = std::complex<std::uint64_t>(
                std::uint64_t(f1.get()), std::uint64_t(f2.get()));
            std::cout << "Inside dataflow(shared) : " << cmplx << std::endl;
            return cmplx;
        },
        fs1, fs2);
    fds.get();

    std::cout << "============================" << std::endl;
    std::cout << "Complete" << std::endl;
    std::cout << "============================" << std::endl << std::endl;
    return 0;
}

struct dummy_tag
{
};

namespace hpx { namespace threads { namespace executors {
    template <>
    struct HPX_EXPORT pool_numa_hint<dummy_tag>
    {
        int operator()(const int, const double, const char*) const
        {
            std::cout << "Hint 1 \n";
            return 1;
        }
        int operator()(const int) const
        {
            std::cout << "Hint 2 \n";
            return 2;
        }
        int operator()(const util::tuple<future<int>, future<double>>&) const
        {
            std::cout << "Hint 3(a) \n";
            return 3;
        }
        int operator()(
            const util::tuple<future<std::uint64_t>, shared_future<float>>&)
            const
        {
            std::cout << "Hint 3(b) \n";
            return 3;
        }
        int operator()(const std::uint16_t, const double) const
        {
            std::cout << "Hint 4(a) \n";
            return 4;
        }
        int operator()(const std::uint32_t, const double&) const
        {
            std::cout << "Hint 4(b) \n";
            return 4;
        }
    };
}}}    // namespace hpx::threads::executors

int hpx_main()
{
    test_async_executor exec;
    test("Testing async custom executor", exec);

    typedef hpx::threads::executors::pool_numa_hint<dummy_tag> dummy_hint;
    hpx::threads::executors::guided_pool_executor<dummy_hint> exec2("default");
    test("Testing guided_pool_executor<dummy_hint>", exec2);

    hpx::threads::executors::guided_pool_executor_shim<dummy_hint> exec3(
        true, "default");
    test("Testing guided_pool_executor_shim<dummy_hint>", exec3);

    return hpx::finalize(0);
}

int main(int argc, char** argv)
{
    int pool_threads = 1;

    // Create the resource partitioner
    hpx::resource::partitioner rp(argc, argv);

    // declare the high priority scheduler type we'll use
    using high_priority_sched =
        hpx::threads::policies::shared_priority_queue_scheduler<>;
    using hpx::threads::policies::scheduler_mode;
    // setup the default pool with our custom priority scheduler
    rp.create_thread_pool("custom",
        [](hpx::threads::thread_pool_init_parameters init,
            hpx::threads::policies::thread_queue_init_parameters
                thread_queue_init)
            -> std::unique_ptr<hpx::threads::thread_pool_base> {
            high_priority_sched::init_parameter_type scheduler_init(
                init.num_threads_, {6, 6, 64}, init.affinity_data_,
                thread_queue_init, "shared-priority-scheduler");
            std::unique_ptr<high_priority_sched> scheduler(
                new high_priority_sched(scheduler_init));

            init.mode_ = scheduler_mode(scheduler_mode::do_background_work |
                scheduler_mode::delay_exit);

            std::unique_ptr<hpx::threads::thread_pool_base> pool(
                new hpx::threads::detail::scheduled_thread_pool<
                    high_priority_sched>(std::move(scheduler), init));
            return pool;
        });

    std::cout << "[main] "
              << "thread_pools created \n";

    // add N cores to mpi pool
    int count = 0;
    for (const hpx::resource::numa_domain& d : rp.numa_domains())
    {
        for (const hpx::resource::core& c : d.cores())
        {
            for (const hpx::resource::pu& p : c.pus())
            {
                if (count < pool_threads)
                {
                    std::cout << "Added pu " << count++ << " to mpi pool\n";
                    rp.add_resource(p, "custom");
                }
            }
        }
    }

    // rp.add_resource(rp.numa_domains()[0].cores()[1], "mpi");
    std::cout << "[main] "
              << "resources added to thread_pools \n";
    return hpx::init(argc, argv);
}
